// vim: ts=4:sw=4
package de.steg0.bi.jr;

import com.lowagie.text.Phrase;
import com.lowagie.text.pdf.BidiLine;
import com.lowagie.text.pdf.ColumnText;
import com.lowagie.text.pdf.PdfContentByte;
import java.util.logging.Logger;

/**
 * Extends {@link ColumnText} to provide access to certain internal
 * fields.
 *
 * <p>This is an experimental implementation which <em>could</em>, if used in
 * place of <code>ColumnText</code> inside {@link
 * JRiTextMeasurer},
 * make the current recursive fit algorithm obsolete. That algorithm is somewhat
 * slow due to its nature, but the problem is that the official iText API
 * does not expose enough information to <em>not</em> make repeated
 * <code>go</code> calls necessary.</p>
 *
 * <p>This class offers a method to obtain the last known good position after a
 * failed fit attempt ({@link #getLastCharacterPosition()}),
 * so that the position that can be reported back to
 * Jasper Reports is immediately known (maybe after an additional word
 * boundary scan).</p>
 *
 * <p>The approach taken here tries to use the internal state of <code>BidiLine</code>
 * to find out about that position. Unfortunately, <code>BidiLine</code> does
 * not keep track of such information in a useful way, which is why a
 * subclass is used that overrides <code>getParagraph</code>. This means
 * that the implementation is dependent of internal structures of iText, which
 * is bad.</p>
 *
 * <p>Another problem is that the approach taken here seems to introduce
 * certain errors. I have observed cases where the result had a deviation of
 * a few characters from the optimum. I have not tracked that down so far.</p>
 *
 * <p>This file is available under the LGPL.</p>
 */
class ColumnText0 
extends ColumnText
{
    private static final Logger LOG=Logger.getLogger("JRiTextMeasurer");

    public ColumnText0(PdfContentByte canvas)
    {
        super(canvas);
    }

    /**
     *
     */
    public void addText(Phrase phrase)
    {
        /* No kidding -- two calls are necessary so that bidiLine is actually
         * set after the call, otherwise the phrase ends up in the
         * "waiting" position */
        super.addText(phrase);
        super.addText(new Phrase());
        this.bidiLine=new BidiLine0(this.bidiLine);
    }

    public int getLastCharacterPosition()
    {
        return ((BidiLine0)this.bidiLine).getParagraphStartPosition() +
               ((BidiLine0)this.bidiLine).getCurrentCharInParagraph();
    }

    /**
     * Extends {@link BidiLine} to access {@link BidiLine#currentChar}, and intercept
     * calls to {@link #getParagraph(int)}.
     */
    public class BidiLine0
    extends BidiLine
    {
        protected int paragraphStartPosition=0;

        /**
         * We can't just override the standard constructor and then let the
         * caller add some chunks, because <code>PdfChunk</code> is non-public
         * and I don't want to fake the package. *sigh*
         *
         * @param bl
         */
        public BidiLine0(BidiLine bl)
        {
            super(bl);
        }

        public boolean getParagraph(int runDirection)
        {
            this.paragraphStartPosition+=this.currentChar;
            LOG.finest("getParagraph(): currentCharBase now: "+this.paragraphStartPosition);
            return super.getParagraph(runDirection);
        }

        public int getParagraphStartPosition()
        {
            LOG.finest("getParagraphStartPosition(): "+this.paragraphStartPosition);
            return this.paragraphStartPosition;
        }

        public int getCurrentCharInParagraph()
        {
            return this.currentChar;
        }
    }

}
