// vim: ts=4:sw=4
package de.steg0.bi.jr;

import com.lowagie.text.Chunk;
import com.lowagie.text.Phrase;
import java.io.ByteArrayInputStream;
import java.text.AttributedCharacterIterator;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRExporterParameter;
import net.sf.jasperreports.engine.JasperCompileManager;
import net.sf.jasperreports.engine.JasperFillManager;
import net.sf.jasperreports.engine.JasperPrint;
import net.sf.jasperreports.engine.JasperReport;
import net.sf.jasperreports.engine.export.JRPdfExporter;
import net.sf.jasperreports.engine.export.JRPdfExporter.NullOutputStream;
import net.sf.jasperreports.engine.util.JRStyledText;

/**
 * Utility functions. This subclasses {@link JRPdfExporter} to provide
 * access to certain methods. Not perfect, but it's arguably better than
 * repeating the code here.
 *
 * <p>This file is available under the LGPL.</p>
 */
public class JRPdfExporterUtil extends JRPdfExporter
{
    private static final Logger LOG=Logger.getLogger("JRPdfExporterUtil");

    /**
     * Initializes the instance, and builds an internal {@link JRPdfExporter}
     * instance used for reading and processing exporter parameters.
     * @param exporterParameters all font and line break related exporter
     * parameters should be provided in this map, in the same way as they
     * are later used in the actual export. The map is not copied,
     * so the caller should provide a copied instance to avoid
     * side-effects.
     * @throws JRException if an error happens initializing the internal
     * exporter instance
     */
    public JRPdfExporterUtil(Map exporterParameters)
    throws JRException
    {
        /*
         * JRPdfExporter's fontMap is private and only initialized
         * when something is exported, hence we need a dummy report
         */
        final JasperReport r=JasperCompileManager.compileReport(new
                ByteArrayInputStream(
                    ("<?xml version=\"1.0\"?>\n"+
                     "<!DOCTYPE jasperReport PUBLIC "+
                     "'-//JasperReports//DTD Report Design//EN' \n"+
                     "'http://jasperreports.sourceforge.net/dtds/"+
                     "jasperreport.dtd'>"+
                     "<jasperReport name=\"dummyReport\">\n"+
                     "</jasperReport>").getBytes()));
        final JasperPrint p=JasperFillManager.fillReport(r,new HashMap());
        this.setParameters(exporterParameters);
        this.setParameter(JRExporterParameter.JASPER_PRINT,p);
        this.setParameter(JRExporterParameter.OUTPUT_STREAM,
                new NullOutputStream());

        this.exportReport();

        /*
         * Now the private data structures of our superclass are
         * initialized, we're safe to use getChunk() and getFont()
         */
    }

    protected Phrase getPhrase(JRStyledText styledText,int start,int end)
    throws JRException
    {
        Phrase phrase = new Phrase();

        String text = styledText.getText();

        int runLimit = 0;

        AttributedCharacterIterator iterator = styledText.getAttributedString()
            .getIterator(null,start,end);

        LOG.finest("getPhrase(): start="+start+", end="+end+
        		", iterator.getRunLimit()="+iterator.getRunLimit());

        while(runLimit < end && (runLimit = iterator.getRunLimit()) <= end)
        {
            Chunk chunk = this.getChunk(iterator.getAttributes(),
                    text.substring(iterator.getIndex(),runLimit));
            LOG.finest("runLimit="+runLimit+"; chunk="+chunk);
            phrase.add(chunk);

            iterator.setIndex(runLimit);
        }

        LOG.finest("getPhrase(): result=\""+phrase.getContent()+"\"");
        return phrase;
    }

}
